# generated by datamodel-codegen:
#   filename:  embedding.yaml

from __future__ import annotations

from typing import List, Union

from pydantic import BaseModel, Field


class NPScoreEmbeddingOutput(BaseModel):
    text: str = Field(..., description='The text that was used to evaluate the similarity')
    distance: Union[float, int] = Field(..., description='Quantile or neighbor distance')
    similarity: float = Field(..., description='What is the similarity of the example to the explanation')


class NPActivation(BaseModel):
    tokens: List[str] = Field(..., description='List of tokens for this text', example=['The', 'cat', 'sat'])
    values: List[float] = Field(
        ..., description='Activation values corresponding to each token', example=[0.5, 0.8, 0.2]
    )


class NPScoreEmbeddingResponse(BaseModel):
    score: float = Field(..., description='The score from 0 to 1')
    breakdown: List[NPScoreEmbeddingOutput] = Field(..., description='Detailed breakdown of the embedding outputs')


class NPScoreEmbeddingRequest(BaseModel):
    activations: List[NPActivation] = Field(..., description='List of activation records to analyze')
    explanation: str = Field(
        ..., description='The explanation to evaluate', example='This neuron activates on references to feline behavior'
    )
    secret: str = Field(..., description='Authentication secret for the API', example='your-secret-key')



from dataclasses import dataclass
from typing import List, Optional

from neuronpedia.np_activation import Activation
from neuronpedia.requests.base_request import NPRequest


@dataclass
class SourceActivationResult:
    modelId: str
    layer: str
    index: str
    values: List[float]


@dataclass
class SourceActivationRequest(NPRequest):
    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        super().__init__("search-all", api_key=api_key)

    def _get_source_set_from_source(self, source: str | list[str]) -> str:
        # if source is a list, return the source set name for the first source, ensuring that all sources have the same source set name
        if isinstance(source, list):
            source_set_name = source[0].split("-", 1)[1]
            for s in source:
                if s.split("-", 1)[1] != source_set_name:
                    raise ValueError(
                        f"All sources must have the same source set name, got {source_set_name} and {s.split('-', 1)[1]}"
                    )
            return source_set_name
        # if source is a string, return the source set name for the source
        return source.split("-", 1)[1]

    def compute_top_features_for_text(
        self,
        model_id: str,
        source: str | list[str],
        text: str,
        ignore_bos: bool = True,
        sort_results_by_activations_at_token_indexes: List[int] = [],
        density_threshold: float = -1,
        num_results: int = 20,
    ) -> list[Activation]:
        payload = {
            "text": text,
            "modelId": model_id,
            "sourceSet": self._get_source_set_from_source(source),
            "selectedLayers": [source] if isinstance(source, str) else source,
            "ignoreBos": ignore_bos,
            "sortIndexes": sort_results_by_activations_at_token_indexes,
            "densityThreshold": density_threshold,
            "numResults": num_results,
        }

        data = self.send_request(method="POST", json=payload, uri="")

        feature_results = [
            SourceActivationResult(
                modelId=item["modelId"],
                layer=item["layer"],
                index=item["index"],
                values=item["values"],
            )
            for item in data["result"]
        ]

        tokens = data["tokens"]

        return [
            Activation(
                modelId=result.modelId,
                source=result.layer,
                index=result.index,
                tokens=tokens,
                values=result.values,
            )
            for result in feature_results
        ]

    def compute_top_features_for_texts(
        self,
        model_id: str,
        source: str | list[str],
        texts: List[str],
        ignore_bos: bool = True,
        sort_results_by_activations_at_token_indexes: List[int] = [],
        density_threshold: float = -1,
        num_results: int = 20,
    ) -> list[list[Activation]]:
        payload = {
            "text": texts,
            "modelId": model_id,
            "sourceSet": self._get_source_set_from_source(source),
            "selectedLayers": [source] if isinstance(source, str) else source,
            "ignoreBos": ignore_bos,
            "sortIndexes": sort_results_by_activations_at_token_indexes,
            "densityThreshold": density_threshold,
            "numResults": num_results,
        }

        data = self.send_request(method="POST", json=payload, uri="")

        # response is a list of results under "results"
        results = data["results"]

        # iterate through each result to create array of SourceActivationResult for each result
        # each result has its own result array
        batch_feature_results = []
        for result in results:
            batch_feature_result = {"tokens": result["tokens"], "results": []}
            for item in result["result"]:
                batch_feature_result["results"].append(
                    SourceActivationResult(
                        modelId=item["modelId"],
                        layer=item["layer"],
                        index=item["index"],
                        values=item["values"],
                    )
                )
            batch_feature_results.append(batch_feature_result)

        # now return the array of array of activations, one for each prompt, in the same order as the input prompts
        batch_activations = []
        for batch_feature_result in batch_feature_results:
            batch_activations.append(
                [
                    Activation(
                        modelId=result.modelId,
                        source=result.layer,
                        index=result.index,
                        tokens=batch_feature_result["tokens"],
                        values=result.values,
                    )
                    for result in batch_feature_result["results"]
                ]
            )

        return batch_activations



from typing import List, Optional
from dataclasses import dataclass

from neuronpedia.requests.base_request import NPRequest
from neuronpedia.types.common.feature import NPFeature


@dataclass
class FeatureWithTopActivation:
    feature: NPFeature
    activationValue: float


@dataclass
class TopKByTokenResult:
    position: int
    token: str
    topFeatures: List[FeatureWithTopActivation]


@dataclass
class TopKByTokenRequest(NPRequest):
    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        super().__init__("search-topk-by-token", api_key=api_key)

    def compute_top_features_by_token_for_text(
        self,
        model_id: str,
        source: str,
        text: str,
        num_results: int = 5,
        ignore_bos: bool = True,
        density_threshold: float = 0.0,
    ) -> List[TopKByTokenResult]:
        payload = {
            "modelId": model_id,
            "source": source,
            "text": text,
            "numResults": num_results,
            "ignoreBos": ignore_bos,
            "densityThreshold": density_threshold,
        }

        data = self.send_request(method="POST", json=payload, uri="")

        token_results = [
            TopKByTokenResult(
                position=item["position"],
                token=item["token"],
                topFeatures=[
                    FeatureWithTopActivation(
                        feature=NPFeature(model=model_id, source=source, index=feature["featureIndex"]),
                        activationValue=feature["activationValue"],
                    )
                    for feature in item["topFeatures"]
                ],
            )
            for item in data["results"]
        ]

        return token_results

    def compute_top_features_by_token_for_texts(
        self,
        model_id: str,
        source: str,
        texts: List[str],
        num_results: int = 10,
        ignore_bos: bool = True,
        density_threshold: float = 0.0,
    ) -> List[List[TopKByTokenResult]]:
        payload = {
            "modelId": model_id,
            "source": source,
            "text": texts,
            "numResults": num_results,
            "ignoreBos": ignore_bos,
            "densityThreshold": density_threshold,
        }

        data = self.send_request(method="POST", json=payload, uri="")

        results = data["results"]

        return [
            [
                TopKByTokenResult(
                    position=item["position"],
                    token=item["token"],
                    topFeatures=[
                        FeatureWithTopActivation(
                            feature=NPFeature(model=model_id, source=source, index=feature["featureIndex"]),
                            activationValue=feature["activationValue"],
                        )
                        for feature in item["topFeatures"]
                    ],
                )
                for item in result["results"]
            ]
            for result in results
        ]



from dataclasses import dataclass, field

from neuronpedia.np_source import Source


@dataclass
class SourceSet:
    name: str
    model_id: str
    description: str | None = None
    url: str | None = None
    sources: list[Source] = field(default_factory=list)

    @classmethod
    def new(
        cls,
        name: str,
        model_id: str,
        description: str | None = None,
        url: str | None = None,
    ) -> "SourceSet":
        from neuronpedia.requests.source_set_request import SourceSetRequest

        return SourceSetRequest().new(name, model_id, description, url)

    @classmethod
    def get(cls, model_id: str, name: str) -> "SourceSet":
        from neuronpedia.requests.source_set_request import SourceSetRequest

        return SourceSetRequest().get(model_id, name)

    def get_source_for_layer_number(self, layer_number: int) -> Source:
        for source in self.sources:
            if source.id == str(layer_number) + "-" + self.name:
                return source
        raise ValueError(
            f"Source for layer {layer_number} not found. Does this model have that many layers?"
        )

    @classmethod
    def from_np_db_json(cls, json_data: dict) -> "SourceSet":
        return cls(
            name=json_data["name"],
            model_id=json_data["modelId"],
            description=json_data["description"],
            url=json_data["urls"][0] if json_data["urls"] else None,
            sources=[
                Source(
                    id=source["id"],
                    model_id=source["modelId"],
                    set_name=source["setName"],
                )
                for source in json_data["sources"]
            ],
        )

    def open_in_browser(self):
        import webbrowser

        webbrowser.open(f"https://neuronpedia.org/{self.model_id}/{self.name}")



"""
Example demonstrating the new context-based API key management using contextvars.
This provides a much cleaner API compared to passing api_key to every constructor.
"""

import neuronpedia


def example_context_manager():
    """Using context manager (recommended for temporary scope)"""
    print("=== Context Manager Example ===")

    with neuronpedia.api_key("your-api-key-here"):
        # All request classes within this context automatically use the API key
        feature_request = neuronpedia.FeatureRequest()
        model_request = neuronpedia.ModelRequest()
        graph_request = neuronpedia.GraphRequest()

        print("✅ All request classes created with context API key")
        # You can create as many request instances as needed without repeating the key

    # Outside the context, the API key is no longer set
    print("Context exited - API key no longer in scope")


def example_global_setting():
    """Using global context setting (good for application-wide usage)"""
    print("\n=== Global Setting Example ===")

    neuronpedia.set_api_key("your-api-key-here")

    # Now all request classes will use this API key
    feature_request = neuronpedia.FeatureRequest()
    vector_request = neuronpedia.VectorRequest()
    steer_request = neuronpedia.SteerChatRequest()

    print("✅ All request classes created with global context API key")


def example_mixed_usage():
    """Demonstrating mixed usage - explicit parameter overrides context"""
    print("\n=== Mixed Usage Example ===")

    neuronpedia.set_api_key("global-api-key")

    # This uses the global context key
    feature_request1 = neuronpedia.FeatureRequest()

    # This overrides with explicit parameter
    feature_request2 = neuronpedia.FeatureRequest(api_key="explicit-api-key")

    # This uses context manager, temporarily overriding global setting
    with neuronpedia.api_key("context-api-key"):
        feature_request3 = neuronpedia.FeatureRequest()

        # Even explicit parameter still works within context
        feature_request4 = neuronpedia.FeatureRequest(api_key="another-explicit-key")

    print("✅ Mixed usage - explicit parameters override context")


def example_nested_contexts():
    """Demonstrating nested context managers"""
    print("\n=== Nested Contexts Example ===")

    with neuronpedia.api_key("outer-key"):
        feature_request1 = neuronpedia.FeatureRequest()
        print("Outer context: created request with outer-key")

        with neuronpedia.api_key("inner-key"):
            feature_request2 = neuronpedia.FeatureRequest()
            print("Inner context: created request with inner-key")

        # Back to outer context
        feature_request3 = neuronpedia.FeatureRequest()
        print("Back to outer context: created request with outer-key")

    print("✅ Nested contexts work correctly")


if __name__ == "__main__":
    print("Demonstrating context-based API key management...")
    print("Note: Replace 'your-api-key-here' with actual API keys to test")

    # example_context_manager()
    # example_global_setting()
    # example_mixed_usage()
    # example_nested_contexts()

    print("\nThis approach is much cleaner than:")
    print("  feature_request = FeatureRequest(api_key='key')")
    print("  model_request = ModelRequest(api_key='key')")
    print("  graph_request = GraphRequest(api_key='key')")
    print("  # ... repeating for every single request class")



from dataclasses import dataclass


@dataclass
class Model:
    id: str
    layers: int
    display_name: str | None = None
    url: str | None = None

    def open_in_browser(self):
        import webbrowser

        webbrowser.open(f"https://neuronpedia.org/{self.id}")

    @classmethod
    def new(
        cls,
        id: str,
        layers: int,
        display_name: str | None = None,
        url: str | None = None,
    ) -> "Model":
        from neuronpedia.requests.model_request import ModelRequest

        return ModelRequest().new(id, layers, display_name, url)



# generated by datamodel-codegen:
#   filename:  topk_by_token.yaml

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field


class NPActivationTopkByTokenRequest(BaseModel):
    secret: str = Field(..., description='API secret for the inference server')
    prompt: str = Field(..., description='Input text prompt to get activations for')
    model: str = Field(..., description='Name of the model to test activations on')
    source: str = Field(
        ...,
        description='Source identifier - could be an SAE ID (eg 5-gemmascope-res-16k). Must be specified with "index", or not at NPActivationAllRequest.',
    )
    top_k: Optional[int] = Field(None, description='The number of features to include for each token position.')
    ignore_bos: bool = Field(
        ..., description='Whether or not to include features whose highest activation value is the BOS token.'
    )


class TopFeature(BaseModel):
    feature_index: int = Field(..., description='The index of the feature in the SAE.')
    activation_value: float = Field(..., description='The activation value of this feature at this token position.')


class NPActivationTopkByTokenResult(BaseModel):
    token_position: int = Field(..., description='The index of the token in the prompt.')
    token: str = Field(..., description='The token string')
    top_features: List[TopFeature]


class NPActivationTopkByTokenResponse(BaseModel):
    results: List[NPActivationTopkByTokenResult]
    tokens: List[str]



# generated by datamodel-codegen:
#   filename:  single.yaml

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field


class NPActivationSingleRequest(BaseModel):
    secret: str = Field(..., description='API secret for the inference server')
    prompt: str = Field(..., description='Input text prompt to get activations for')
    model: str = Field(..., description='Name of the model to test activations on')
    source: Optional[str] = Field(
        None,
        description='Source identifier - could be an SAE ID (eg 5-gemmascope-res-16k). Must be specified with "index", or not at all.',
    )
    index: Optional[str] = Field(None, description='Index of the SAE. Must be specified with "source", or not at all.')
    vector: Optional[List[float]] = Field(
        None, description='Custom vector to test activations. Must be specified with "hook".'
    )
    hook: Optional[str] = Field(
        None, description='Hook that the custom vector applies to. Must be specified with "vector".'
    )


class NPActivationSingleActivation(BaseModel):
    values: List[float]
    max_value: float
    max_value_index: int
    dfa_values: Optional[List[float]] = None
    dfa_max_value: Optional[float] = None
    dfa_target_index: Optional[int] = None


class NPActivationSingleResponse(BaseModel):
    activation: NPActivationSingleActivation
    tokens: List[str]



from dataclasses import dataclass


@dataclass
class SAEFeature:
    modelId: str
    source: str
    index: str
    jsonData: str

    @classmethod
    def get(cls, model_id: str, source: str, index: str) -> "SAEFeature":

        from neuronpedia.requests.sae_feature_request import SAEFeatureRequest

        request = SAEFeatureRequest()
        return request.get(model_id, source, index)



from typing import Optional

from neuronpedia.np_source_set import SourceSet
from neuronpedia.requests.base_request import NPRequest


class SourceSetRequest(NPRequest):
    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        super().__init__("source-set", api_key=api_key)

    def new(
        self,
        name: str,
        model_id: str,
        description: str | None = None,
        url: str | None = None,
    ) -> SourceSet:
        """
        Create a new source set on Neuronpedia.
        Before uploading features, you must create a source set, which is a collection of sources.
        Each *source* is a set of features for a layer of a model. For example, one sparse autoencoder, which can contain thousands of features, is one source.
        Sources are named as [layer]-[source_set_name]. For example, "20-gemmascope-res-16k" is the residual stream of the 20th layer, with 16k features. Its source set name is "gemmascope-res-16k".

        When you create a source set, it automatically creates a source for each layer of the model.
        If you create a source set called "puppyscope-res-16k" for a 3 layer "pup-model-3" model, it will create 3 sources: "0-puppyscope-res-16k", "1-puppyscope-res-16k", and "2-puppyscope-res-16k".

        After creating the source set, you can upload features (and activations + explanations) to it, by referencing its source. For the example above, you upload the to layer 1 by calling the Feature.upload_batch() method with modelId: "pup-model-3", source: "1-puppyscope-res-16k", and the feature details for each feature index.
        Only the creator of the source set can upload features to it.

        Args:
            name: Name of the source set (eg "gemmascope-res-16k" - no strict rules, but usually include the author name or abbreviation, plus abbreviation of the hook used)
            model_id: ID of the model this source set belongs to (eg "gemma-2-2b")
            description: Optional description of the source set (eg "Residual Stream - 16k")
            url: Optional URL associated with the source set (eg "https://huggingface.co/google/gemma-scope-2b-pt-res/tree/main")

        Returns:
            SourceSet: Response containing the created source set information
        """
        payload = {
            "name": name,
            "modelId": model_id,
            "description": description,
            "url": url,
        }
        response = self.send_request(
            method="POST",
            json=payload,
            uri="new",
        )
        return SourceSet.from_np_db_json(response)

    def get(self, model_id: str, name: str) -> SourceSet:
        response = self.send_request(
            method="GET",
            uri=f"{model_id}/{name}",
        )
        return SourceSet.from_np_db_json(response)



# generated by datamodel-codegen:
#   filename:  feature.yaml

from __future__ import annotations

from pydantic import BaseModel, Field


class NPFeature(BaseModel):
    model: str = Field(..., example='gpt2-small')
    source: str = Field(..., example='0-res-jb')
    index: int = Field(..., example=14057)



# generated by datamodel-codegen:
#   filename:  fuzz_detection.yaml

from __future__ import annotations

from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel, Field


class NPScoreFuzzDetectionType(Enum):
    fuzz = 'fuzz'
    detection = 'detection'


class NPScoreClassifierOutput(BaseModel):
    str_tokens: Optional[List[str]] = Field(None, description='List of strings')
    activations: Optional[List[float]] = Field(None, description='List of floats')
    distance: Optional[Union[float, int]] = Field(None, description='Quantile or neighbor distance')
    ground_truth: Optional[bool] = Field(None, description='Whether the example is activating or not')
    prediction: Optional[bool] = Field(False, description='Whether the model predicted the example activating or not')
    highlighted: Optional[bool] = Field(False, description='Whether the sample is highlighted')
    probability: Optional[float] = Field(0.0, description='The probability of the example activating')
    correct: Optional[bool] = Field(False, description='Whether the prediction is correct')


class NPActivation(BaseModel):
    tokens: List[str] = Field(..., description='List of tokens for this text', example=['The', 'cat', 'sat'])
    values: List[float] = Field(
        ..., description='Activation values corresponding to each token', example=[0.5, 0.8, 0.2]
    )


class NPScoreFuzzDetectionResponse(BaseModel):
    score: float = Field(..., description='The score from 0 to 1')
    breakdown: List[NPScoreClassifierOutput] = Field(
        ..., description='Detailed breakdown of the classification outputs'
    )


class NPScoreFuzzDetectionRequest(BaseModel):
    activations: List[NPActivation] = Field(..., description='List of activation records to analyze')
    explanation: str = Field(
        ..., description='The explanation to evaluate', example='This neuron activates on references to feline behavior'
    )
    openrouter_key: str = Field(..., description='API key for OpenRouter service', example='sk-or-v1-...')
    model: str = Field(..., description='Model identifier to use for scoring', example='openai/gpt-4o-mini')
    type: NPScoreFuzzDetectionType = Field(..., description='Type of scoring to perform - either fuzzing or detection')
    secret: str = Field(..., description='Authentication secret for the API', example='your-secret-key')



from typing import Optional

from neuronpedia.np_list import NPList, NPListItem
from neuronpedia.requests.base_request import NPRequest


class ListRequest(NPRequest):
    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        super().__init__("list", api_key=api_key)

    def get_owned(self) -> list[NPList]:
        response = self.send_request(method="POST", uri="list")

        return [
            NPList(
                id=list["id"],
                name=list["name"],
                description=list["description"],
                items=[],
            )
            for list in response
        ]

    def new(self, name: str, description: str = "") -> NPList:
        payload = {"name": name, "description": description}
        response = self.send_request(method="POST", json=payload, uri="new")
        return NPList(
            id=response["id"],
            name=response["name"],
            description=response["description"],
            items=[],
        )

    def add_items(self, nplist: NPList, items: list[NPListItem]) -> NPList:
        formatted_items = [
            {
                "modelId": item.model_id,
                "layer": item.source,
                "index": item.index,
                "description": item.description,
            }
            for item in items
        ]
        payload = {"listId": nplist.id, "featuresToAdd": formatted_items}
        response = self.send_request(method="POST", json=payload, uri="add-features")
        for item in response:
            nplist.items.append(
                NPListItem(
                    model_id=item["modelId"],
                    source=item["layer"],
                    index=item["index"],
                )
            )
        return nplist

    def get(self, list_id: str) -> NPList:
        payload = {"listId": list_id}
        response = self.send_request(method="POST", json=payload, uri="get")

        return NPList(
            id=response["id"],
            name=response["name"],
            description=response["description"],
            items=[
                NPListItem(
                    model_id=item["modelId"],
                    source=item["layer"],
                    index=item["index"],
                )
                for item in response["neurons"]
            ],
        )



from typing import List, Optional

from neuronpedia.np_activation import Activation
from neuronpedia.requests.base_request import NPRequest
from requests import Response


class ActivationRequest(NPRequest):
    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        super().__init__("activation", api_key=api_key)

    def compute_all_source_activations_for_text(
        self, model_id: str, source: str, text: str | list[str]
    ):
        payload = {
            "customText": text,
            "modelId": model_id,
            "source": source,
        }
        data = self.send_request(method="POST", json=payload, uri="source")

        return data["results"]

    def compute_activation_for_text(
        self, model_id: str, source: str, index: str, text: str
    ) -> Activation:
        if isinstance(index, int):
            index = str(index)
        payload = {
            "feature": {
                "modelId": model_id,
                "layer": source,
                "index": index,
            },
            "customText": text,
        }

        result = self.send_request(method="POST", json=payload, uri="new")

        return Activation(
            modelId=model_id,
            source=source,
            index=index,
            tokens=result["tokens"],
            values=result["values"],
        )

    def compute_activation_for_texts(
        self, model_id: str, source: str, index: str, texts: List[str]
    ) -> List[Activation]:
        if isinstance(index, int):
            index = str(index)
        payload = {
            "feature": {
                "modelId": model_id,
                "layer": source,
                "index": index,
            },
            "customText": texts,
        }

        results = self.send_request(method="POST", json=payload, uri="new")
        return [
            Activation(
                modelId=model_id,
                source=source,
                index=index,
                tokens=result["tokens"],
                values=result["values"],
            )
            for result in results
        ]

    def upload_batch(
        self,
        model_id: str,
        source: str,
        index: str,
        activations: List[Activation],
    ) -> Response:
        payload = {
            "modelId": model_id,
            "source": source,
            "index": index,
            "activations": activations,
        }
        return self.send_request(method="POST", json=payload, uri="upload-batch")



import os
from dataclasses import dataclass
from typing import List


@dataclass
class NPGraphMetadata:
    """A Graph Metadata returned by the Neuronpedia API."""

    id: str  # A unique ID automatically generated by Neuronpedia
    model_id: str  # The model that the graph belongs to
    slug: str  # A user-defined identifier for the graph. Must be globally unique combined with model_id.
    prompt_tokens: List[str]  # The tokens that were used to generate the graph
    prompt: str  # The prompt that was used to generate the graph
    title_prefix: str
    json_url: str  # URL to the actual graph JSON file
    url: str | None = None  # URL to the graph on Neuronpedia
    url_embed: str | None = None  # URL to the embedded graph on Neuronpedia

    def __post_init__(self):
        if not self.url:
            USE_LOCALHOST = os.getenv("USE_LOCALHOST", "false").lower() == "true"
            BASE_URL = (
                "https://neuronpedia.org"
                if not USE_LOCALHOST
                else "http://localhost:3000"
            )
            self.url = f"{BASE_URL}/{self.model_id}/graph?slug={self.slug}"
            self.url_embed = f"{self.url}&embed=true"

    def __eq__(self, other: "NPGraphMetadata") -> bool:
        return (
            self.model_id == other.model_id
            and self.slug == other.slug
            and self.prompt_tokens == other.prompt_tokens
            and self.prompt == other.prompt
            and self.title_prefix == other.title_prefix
            and self.json_url == other.json_url
            and self.url == other.url
        )

    def delete(self):
        # import here to avoid circular import
        from neuronpedia.requests.graph_request import GraphRequest

        return GraphRequest().delete(self)

    @classmethod
    def get(cls, model_id: str, slug: str) -> "NPGraphMetadata":
        # import here to avoid circular import
        from neuronpedia.requests.graph_request import GraphRequest

        return GraphRequest().get(model_id, slug)

    @classmethod
    def list_owned(cls) -> List["NPGraphMetadata"]:
        # import here to avoid circular import
        from neuronpedia.requests.graph_request import GraphRequest

        return GraphRequest().list_owned()

    @classmethod
    def generate(
        cls,
        model_id: str,
        prompt: str,
        graph_id: str,
        max_n_logits: int = 10,
        desired_logit_prob: float = 0.95,
        node_threshold: float = 0.8,
        edge_threshold: float = 0.98,
    ) -> "NPGraphMetadata":
        # import here to avoid circular import
        from neuronpedia.requests.graph_request import GraphRequest

        return GraphRequest().generate(
            model_id,
            prompt,
            graph_id,
            max_n_logits,
            desired_logit_prob,
            node_threshold,
            edge_threshold,
        )

    @classmethod
    def upload(cls, json_str: str) -> "NPGraphMetadata":
        # import here to avoid circular import
        from neuronpedia.requests.graph_request import GraphRequest

        return GraphRequest().upload(json_str)

    @classmethod
    def upload_file(cls, filepath: str) -> "NPGraphMetadata":
        # import here to avoid circular import
        from neuronpedia.requests.graph_request import GraphRequest

        return GraphRequest().upload_file(filepath)



from dataclasses import dataclass
from typing import List


@dataclass
class Activation:
    modelId: str
    source: str
    index: int
    tokens: List[str]
    values: List[float]
    quantileMax: float | None = None
    quantileMin: float | None = None
    quantileFraction: float | None = None

    def __post_init__(self):
        if self.quantileFraction is not None:
            if not (0 <= self.quantileFraction <= 1):
                raise ValueError(f"quantileFraction must be between 0 and 1 inclusive, got {self.quantileFraction}")

    def __str__(self):
        lines = [f"Activation for modelId: {self.modelId}, source: {self.source}, index: {self.index}"]
        for token, value in zip(self.tokens, self.values):
            lines.append(f"{token}\t{value}")
        return "\n".join(lines)

    # converts from the Neuronpedia database type
    @classmethod
    def from_np_db_json(cls, json_data: dict) -> "Activation":
        # Ensure tokens and values have the same length
        if len(json_data["tokens"]) != len(json_data["values"]):
            raise ValueError(
                f"Length mismatch: tokens has {len(json_data['tokens'])} elements, values has {len(json_data['values'])} elements"
            )
        return cls(
            modelId=json_data["modelId"],
            source=json_data["source"],
            index=int(json_data["index"]),
            tokens=json_data["tokens"],
            values=json_data["values"],
            quantileMax=json_data["binMax"],
            quantileMin=json_data["binMin"],
            quantileFraction=json_data["binContains"],
        )



from dataclasses import dataclass
from typing import List

from neuronpedia.np_feature import Feature


@dataclass
class Source:
    id: str
    model_id: str
    set_name: str

    def open_in_browser(self):
        import webbrowser

        webbrowser.open(f"https://neuronpedia.org/{self.model_id}/{self.id}")

    def upload_batch(self, features: List[Feature]):
        from neuronpedia.requests.feature_request import FeatureRequest

        request = FeatureRequest()
        return request.upload_batch(self.model_id, self.id, features)



"""
Neuronpedia Python Client Library

This library provides a clean interface to interact with Neuronpedia's API.

API Key Configuration:
There are three ways to configure your API key:

1. Context manager (recommended):
   ```python
   import neuronpedia

   with neuronpedia.api_key("your-api-key"):
       feature_request = neuronpedia.FeatureRequest()
   ```

2. Global context setting:
   ```python
   import neuronpedia

   neuronpedia.set_api_key("your-api-key")
   feature_request = neuronpedia.FeatureRequest()
   ```

3. Direct parameter (for explicit control):
   ```python
   feature_request = neuronpedia.FeatureRequest(api_key="your-api-key")
   ```

4. Environment variable (existing behavior):
   ```python
   # Set NEURONPEDIA_API_KEY environment variable
   feature_request = neuronpedia.FeatureRequest()
   ```
"""

import contextvars
from contextlib import contextmanager
from typing import Optional

# Context variable for API key
_api_key_context: contextvars.ContextVar[Optional[str]] = contextvars.ContextVar(
    "neuronpedia_api_key", default=None
)


def set_api_key(api_key: str) -> None:
    """
    Set the API key in the current context.

    Args:
        api_key: Your Neuronpedia API key
    """
    _api_key_context.set(api_key)


def get_api_key() -> Optional[str]:
    """
    Get the API key from the current context.

    Returns:
        The API key if set in context, None otherwise
    """
    return _api_key_context.get()


@contextmanager
def api_key(api_key: str):
    """
    Context manager for temporarily setting an API key.

    Args:
        api_key: Your Neuronpedia API key

    Example:
        ```python
        with neuronpedia.api_key("your-key"):
            feature_request = FeatureRequest()
        ```
    """
    token = _api_key_context.set(api_key)
    try:
        yield
    finally:
        _api_key_context.reset(token)


from neuronpedia.requests.activation_request import ActivationRequest

# Import request classes for easy access
from neuronpedia.requests.feature_request import FeatureRequest
from neuronpedia.requests.graph_request import GraphRequest
from neuronpedia.requests.list_request import ListRequest
from neuronpedia.requests.model_request import ModelRequest
from neuronpedia.requests.sae_feature_request import SAEFeatureRequest
from neuronpedia.requests.source_set_request import SourceSetRequest
from neuronpedia.requests.steer_request import SteerChatRequest, SteerCompletionRequest
from neuronpedia.requests.vector_request import VectorRequest

__all__ = [
    "set_api_key",
    "get_api_key",
    "api_key",
    "FeatureRequest",
    "ModelRequest",
    "GraphRequest",
    "SourceSetRequest",
    "VectorRequest",
    "ListRequest",
    "SAEFeatureRequest",
    "SteerChatRequest",
    "SteerCompletionRequest",
    "ActivationRequest",
]



from typing import List, Optional

from neuronpedia.np_activation import Activation
from neuronpedia.np_explanation import Explanation
from neuronpedia.np_feature import Feature
from neuronpedia.requests.base_request import NPRequest

MAX_FEATURES_PER_BATCH = 128


class FeatureRequest(NPRequest):
    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        super().__init__("feature", api_key=api_key)

    def get(self, model_id: str, source: str, index: int) -> Feature:
        result = self.send_request(method="GET", uri=f"{model_id}/{source}/{index}")

        explanations = [
            Explanation.from_np_db_json(explanation)
            for explanation in result["explanations"]
        ]

        activations = [
            Activation.from_np_db_json(activation)
            for activation in result["activations"]
        ]

        return Feature(
            modelId=result["modelId"],
            source=result["layer"],
            index=result["index"],
            density=result["frac_nonzero"],
            explanations=explanations,
            activations=activations,
        )

    def upload_batch(self, model_id: str, source: str, features: List[Feature]):
        if len(features) > MAX_FEATURES_PER_BATCH:
            raise ValueError(
                f"Cannot upload more than {MAX_FEATURES_PER_BATCH} features at a time"
            )
        if len(features) == 0:
            raise ValueError("Cannot upload 0 features")

        # make the features payload
        features_payload = []
        for feature in features:
            to_append = {
                "index": feature.index,
                "density": feature.density,
                "activations": [],
                "explanations": [],
                "topLogits": [
                    {"token": logit.token, "value": logit.value}
                    for logit in feature.top_logits
                ]
                if feature.top_logits is not None
                else None,
                "bottomLogits": [
                    {"token": logit.token, "value": logit.value}
                    for logit in feature.bottom_logits
                ]
                if feature.bottom_logits is not None
                else None,
            }
            if feature.explanations is not None:
                for explanation in feature.explanations:
                    to_append["explanations"].append(
                        {
                            "text": explanation.text,
                            "methodName": explanation.method,
                            "modelName": explanation.explainer_model,
                        }
                    )
            if feature.activations is not None:
                for activation in feature.activations:
                    to_append["activations"].append(
                        {
                            "tokens": activation.tokens,
                            "values": activation.values,
                            "quantileMax": activation.quantileMax,
                            "quantileMin": activation.quantileMin,
                            "quantileFraction": activation.quantileFraction,
                        }
                    )
            features_payload.append(to_append)
        payload = {
            "modelId": model_id,
            "source": source,
            "features": features_payload,
        }
        return self.send_request(method="POST", uri="upload-batch", json=payload)



# generated by datamodel-codegen:
#   filename:  steer.yaml

from __future__ import annotations

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Field, confloat, conint


class NPFeature(BaseModel):
    model: str = Field(..., example='gpt2-small')
    source: str = Field(..., example='0-res-jb')
    index: int = Field(..., example=14057)


class NPSteerVector(BaseModel):
    steering_vector: List[float]
    strength: float
    hook: str = Field(..., example='blocks.0.hook_resid_pre')


class NPSteerType(Enum):
    STEERED = 'STEERED'
    DEFAULT = 'DEFAULT'


class NPSteeringMethod(Enum):
    simple_additive = 'simple_additive'
    orthogonal_decomp = 'orthogonal_decomp'


class NPSteerCompletionResponse(BaseModel):
    STEERED: Optional[str] = Field(
        None, description='The full text of the steered prompt and response, concatenated as one string.'
    )
    DEFAULT: Optional[str] = Field(
        None, description='The full text of the default (non-steered) prompt and response, concatenated as one string.'
    )


class NPSteerChatTemplate(BaseModel):
    content: str = Field(..., description='The chat message')
    role: str = Field(..., description='The role of the message (eg "model", "user", etc)')


class NPSteerChatResult(BaseModel):
    chat_template: List[NPSteerChatTemplate]
    raw: str


class NPSteerCompletionChatResponse(BaseModel):
    STEERED: Optional[NPSteerChatResult] = None
    DEFAULT: Optional[NPSteerChatResult] = None
    input: Optional[NPSteerChatResult] = None


class NPSteerFeature(NPFeature):
    strength: float = Field(..., example=5)
    steering_vector: Optional[List[float]] = None


class NPSteerCompletionRequest(BaseModel):
    secret: str = Field(..., description='API secret for the inference server')
    prompt: str = Field(..., description='Text to pass the model for completion')
    model: str = Field(..., description='Name of the model')
    steering_method: NPSteeringMethod
    normalize_steering: bool
    types: List[NPSteerType] = Field(
        ...,
        description='Array that specifies whether or not to generate STEERED output, DEFAULT (non-steered) output, or both.',
        min_items=1,
    )
    features: Optional[List[NPSteerFeature]] = Field(None, description='Features to steer towards or away from')
    vectors: Optional[List[NPSteerVector]] = None
    n_completion_tokens: conint(ge=1) = Field(..., description='Number of completion tokens to generate')
    temperature: confloat(ge=0.0)
    strength_multiplier: float = Field(..., description='The steering strength will be multiplied by this number')
    freq_penalty: float
    seed: float


class NPSteerCompletionChatRequest(NPSteerCompletionRequest):
    steer_special_tokens: bool
    prompt: List[NPSteerChatTemplate] = Field(..., description='Array of chat messages to pass to the model')



from dataclasses import dataclass
from typing import List


@dataclass
class NPListItem:
    model_id: str
    source: str
    index: str
    description: str = ""


@dataclass
class NPList:
    id: str
    name: str
    description: str
    items: List[NPListItem]

    @classmethod
    def new(cls, name: str, description: str = "") -> "NPList":
        # import here to avoid circular import
        from neuronpedia.requests.list_request import ListRequest

        return ListRequest().new(name, description)

    @classmethod
    def get_owned(cls) -> list["NPList"]:
        # import here to avoid circular import
        from neuronpedia.requests.list_request import ListRequest

        return ListRequest().get_owned()

    def add_items(self, items: list[NPListItem]) -> "NPList":
        # import here to avoid circular import
        from neuronpedia.requests.list_request import ListRequest

        return ListRequest().add_items(self, items)

    @classmethod
    def get(cls, list_id: str) -> "NPList":
        # import here to avoid circular import
        from neuronpedia.requests.list_request import ListRequest

        return ListRequest().get(list_id)



from dataclasses import dataclass


@dataclass
class Explanation:
    modelId: str
    source: str
    index: int
    text: str
    method: str | None = None
    explainer_model: str | None = None

    # converts from the Neuronpedia database type
    @classmethod
    def from_np_db_json(cls, json_data: dict) -> "Explanation":
        return cls(
            modelId=json_data["modelId"],
            source=json_data["source"],
            index=int(json_data["index"]),
            text=json_data["description"],
            method=json_data.get("typeName"),
            explainer_model=json_data.get("explanationModelName"),
        )



# generated by datamodel-codegen:
#   filename:  default.yaml

from __future__ import annotations

from typing import List

from pydantic import BaseModel, Field


class NPActivation(BaseModel):
    tokens: List[str] = Field(..., description='List of tokens for this text', example=['The', 'cat', 'sat'])
    values: List[float] = Field(
        ..., description='Activation values corresponding to each token', example=[0.5, 0.8, 0.2]
    )


class NPExplainDefaultResponse(BaseModel):
    explanation: str = Field(..., description='The generated explanation for the given set of activations')


class NPExplainDefaultRequest(BaseModel):
    activations: List[NPActivation] = Field(..., description='List of activation records to analyze')
    openrouter_key: str = Field(..., description='API key for OpenRouter service', example='sk-or-v1-...')
    model: str = Field(
        ..., description='Model identifier to use for explanation generation', example='openai/gpt-4o-mini'
    )
    secret: str = Field(..., description='Authentication secret for the API', example='your-secret-key')



# generated by datamodel-codegen:
#   filename:  sae_eval.yaml

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field


class NPSaeEval(BaseModel):
    id: str = Field(..., description='ID of the SAE Eval')
    typeName: str = Field(..., description='SAE Eval Type - should equal the "name" property of NPSaeEvalType')
    modelId: str = Field(..., description='The model ID of the SAE being evaluated')
    sourceId: str = Field(..., description='The source ID of the SAE being evaluated')
    output: str = Field(
        ...,
        description='The output data, matching the outputSchema of the NPSaeEvalType, with the exception of eval_result_details, which will go under detailedMetrics',
    )
    detailedMetrics: Optional[str] = Field(
        None,
        description='The eval_result_details of the output data. Kept in a separate field/column due to large size.',
    )


class NPSaeEvalType(BaseModel):
    name: str = Field(..., description='ID of the Eval Type')
    displayName: str = Field(..., description='A UI-friendly display name of the eval type')
    description: str = Field(..., description='A brief description of the eval')
    outputSchema: str = Field(
        ...,
        description='JSON output schema used to parse the eval output results. From the SAEBench repo - for example https://github.com/adamkarvonen/SAEBench/blob/main/sae_bench/evals/autointerp/eval_output_schema_autointerp.json',
    )
    url: Optional[str] = Field(None, description='URL to the eval')
    evals: Optional[List[NPSaeEval]] = Field(None, description='The actual eval results of this type')


class NPSaeEvalListRequest(BaseModel):
    pass


class NPSaeEvalListResponse(BaseModel):
    evalTypes: Optional[List[NPSaeEvalType]] = Field(None, description='List of SAE eval types and their results')



import gzip
import json
import os
from typing import List, Optional

import requests
from neuronpedia.np_graph_metadata import NPGraphMetadata
from neuronpedia.requests.base_request import NPRequest
from requests import Response

UPLOAD_FILE_SIZE_LIMIT_BYTES = 100 * 1024 * 1024  # 100MB


class GraphRequest(NPRequest):
    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        super().__init__("graph", api_key=api_key)

    def upload(
        self,
        json_str: str,
    ) -> NPGraphMetadata:
        """
        Upload a JSON graph to Neuronpedia.
        """

        # parse the json string
        loaded_json = json.loads(json_str)

        # TODO: json file checking should be done using the Open Source Circuit Finding repo
        # Check that loaded_json has required fields
        required_fields = ["metadata"]
        for field in required_fields:
            if field not in loaded_json:
                raise ValueError(f"JSON file must contain '{field}' field")

        # Check metadata contains required fields
        if "metadata" in loaded_json:
            required_metadata = [
                "prompt",
                "prompt_tokens",
                "scan",
                "slug",
            ]
            for field in required_metadata:
                if field not in loaded_json["metadata"]:
                    raise ValueError(f"Metadata must contain '{field}' field")

        # we allow any graph to be uploaded
        # if loaded_json["metadata"]["scan"] not in SUPPORTED_GRAPH_MODELS:
        #     raise ValueError(
        #         f"Model {loaded_json['metadata']['scan']} is not supported. Must be one of {SUPPORTED_GRAPH_MODELS}"
        #     )

        model_id = loaded_json["metadata"]["scan"]
        slug = loaded_json["metadata"]["slug"]

        # gzip so we know what the size is
        json_str_gzip = gzip.compress(json_str.encode("utf-8"))

        payload = {
            "filename": slug + ".json",
            "contentLength": len(json_str_gzip),
            "contentType": "application/json",
        }

        response = self.send_request(
            method="POST",
            json=payload,
            uri="signed-put",
        )

        # response will be url and putRequestId
        url = response["url"]
        put_request_id = response["putRequestId"]

        # if either are missing, raise error
        if not url or not put_request_id:
            raise ValueError(
                "Failed to get url or putRequestId from response. Response: {response}"
            )

        print("Upload progress - received signed put request.")

        # upload the file
        response = requests.put(
            url, data=json_str_gzip, headers={"Content-Encoding": "gzip"}
        )

        # check the response is 200
        if response.status_code != 200:
            raise ValueError(f"Failed to upload file to S3. Response: {response.text}")

        print("Upload progress - uploaded file to S3.")

        # now persist it to the database
        response = self.send_request(
            method="POST",
            uri="save-to-db",
            json={"putRequestId": put_request_id},
        )

        print(
            f"Upload to Neuronpedia complete!\nView the graph at the following URL:\n{response['url']}"
        )

        graph_metadata = self.get(model_id, slug)

        return graph_metadata

    def upload_file(
        self,
        filepath: str,
    ) -> NPGraphMetadata:
        """
        Upload a graph file to Neuronpedia.

        Args:
            filepath: Local path to the JSON graph file to upload.

        Returns:
            NPGraphMetadata: The metadata for the uploaded graph. Check its url field to see and use the graph on Neuronpedia, or json_url to get the JSON file.

        Raises:
            FileNotFoundError: If the file doesn't exist at the specified path.
            ValueError: If the file exceeds size limits, isn't valid JSON, or doesn't
                        contain required metadata fields.
        """

        if not os.path.exists(filepath):
            raise FileNotFoundError(f"File not found at path: {filepath}")

        if os.path.getsize(filepath) > UPLOAD_FILE_SIZE_LIMIT_BYTES:
            raise ValueError(
                f"File size exceeds the limit of {UPLOAD_FILE_SIZE_LIMIT_BYTES / 1024 / 1024} megabytes"
            )

        # if not json, raise error
        if not filepath.endswith(".json"):
            raise ValueError("File must be a JSON file")

        # load file as string
        with open(filepath, "r") as f:
            json_str = f.read()

        # upload the json
        graph_metadata = self.upload(json_str)

        return graph_metadata

    def delete(self, graph_metadata: NPGraphMetadata) -> Response:
        return self._delete(graph_metadata.model_id, graph_metadata.slug)

    def _delete(self, model_id: str, slug: str) -> Response:
        payload = {
            "modelId": model_id,
            "slug": slug,
        }
        return self.send_request(
            method="POST",
            uri="delete",
            json=payload,
        )

    def get(self, model_id: str, slug: str) -> NPGraphMetadata:
        response = self.send_request(
            method="GET",
            uri=f"{model_id}/{slug}",
        )
        return NPGraphMetadata(
            id=response["id"],
            model_id=response["modelId"],
            slug=response["slug"],
            prompt_tokens=response["promptTokens"],
            prompt=response["prompt"],
            title_prefix=response["titlePrefix"],
            json_url=response["url"],
        )

    def list_owned(self) -> List[NPGraphMetadata]:
        response = self.send_request(
            method="POST",
            uri="list-owned",
        )

        return [
            NPGraphMetadata(
                id=graph["id"],
                model_id=graph["modelId"],
                slug=graph["slug"],
                prompt_tokens=graph["promptTokens"],
                prompt=graph["prompt"],
                title_prefix=graph["titlePrefix"],
                json_url=graph["url"],
            )
            for graph in response
        ]

    def generate(
        self,
        model_id: str,
        prompt: str,
        slug: str,
        max_n_logits: int = 10,
        desired_logit_prob: float = 0.95,
        node_threshold: float = 0.8,
        edge_threshold: float = 0.98,
    ) -> NPGraphMetadata:
        payload = {
            "modelId": model_id,
            "prompt": prompt,
            "slug": slug,
            "maxNLogits": max_n_logits,
            "desiredLogitProb": desired_logit_prob,
            "nodeThreshold": node_threshold,
            "edgeThreshold": edge_threshold,
        }
        self.send_request(
            method="POST",
            uri="generate",
            json=payload,
        )

        # now get the graph metadata
        graph_metadata = self.get(model_id, slug)
        return graph_metadata



# generated by datamodel-codegen:
#   filename:  util.yaml

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field


class NPGetSaeVectorRequest(BaseModel):
    model: str = Field(..., example='gpt2-small')
    source: str = Field(..., example='0-res-jb')
    index: int = Field(..., example=14057)
    secret: str = Field(..., description='Authentication secret for the API', example='your-secret-key')


class NPGetSaeVectorResponse(BaseModel):
    vector: List[float]


class NPFeature(BaseModel):
    model: str = Field(..., example='gpt2-small')
    source: str = Field(..., example='0-res-jb')
    index: int = Field(..., example=14057)


class TopkDecoderCossimFeature(BaseModel):
    feature: Optional[NPFeature] = None
    cosine_similarity: Optional[float] = None


class NPSaeTopkByDecoderCosSimResponse(BaseModel):
    feature: Optional[NPFeature] = None
    topk_decoder_cossim_features: Optional[List[TopkDecoderCossimFeature]] = None


class NPSaeTopkByDecoderCosSimRequest(BaseModel):
    feature: Optional[NPFeature] = None
    secret: str
    num_results: int



# generated by datamodel-codegen:
#   filename:  all.yaml

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field


class NPActivationAllActivation(BaseModel):
    source: str
    index: int
    values: List[float]
    sum_values: Optional[float] = None
    max_value: Optional[float] = None
    max_value_index: Optional[float] = None
    dfa_values: Optional[List[float]] = None
    dfa_target_index: Optional[int] = None
    dfa_max_value: Optional[float] = None


class NPActivationAllResponse(BaseModel):
    activations: Optional[List[NPActivationAllActivation]] = None
    tokens: Optional[List[str]] = None
    counts: Optional[List[List[float]]] = Field(
        None,
        description='Not currently supported and may be incorrect. This is the number of features that activated by layer, starting from layer 0 of this SAE. Need to be redesigned.',
    )


class NPActivationAllRequest(BaseModel):
    secret: str = Field(..., description='API secret for the inference server')
    prompt: str = Field(..., description='Input text prompt to get activations for')
    model: str = Field(..., description='Name of the model to test activations on')
    source_set: str = Field(..., description='The source set name of the SAEs (eg gemmascope-res-16k)')
    selected_sources: List[str] = Field(
        ...,
        description='List of specific SAEs to get activations for (eg ["0-gemmascope-res-65k", "5-gemmascope-res-65k"]). If not specified, will get activations for all SAEs in the source set.',
    )
    sort_by_token_indexes: List[int] = Field(
        ..., description='Sort the results by the sum of the activations at the specified token indexes.'
    )
    ignore_bos: bool = Field(
        ..., description='Whether or not to include features whose highest activation value is the BOS token.'
    )
    feature_filter: Optional[List[int]] = Field(
        None,
        description='Optional. If specified, will only return features that match the indexes specified. Can only be used if we\'re testing just one SAE ("selected_sources" length = 1).',
    )
    num_results: Optional[int] = Field(25, description='Optional. The number of top features to return.')



# generated by datamodel-codegen:
#   filename:  activation.yaml

from __future__ import annotations

from typing import List

from pydantic import BaseModel, Field


class NPActivation(BaseModel):
    tokens: List[str] = Field(..., description='List of tokens for this text', example=['The', 'cat', 'sat'])
    values: List[float] = Field(
        ..., description='Activation values corresponding to each token', example=[0.5, 0.8, 0.2]
    )



from typing import Optional

from neuronpedia.np_vector import NPVector
from neuronpedia.requests.base_request import NPRequest
from requests import Response

VALID_HOOK_TYPES = ["hook_resid_pre"]


class VectorRequest(NPRequest):
    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        super().__init__("vector", api_key=api_key)

    def new(
        self,
        label: str,
        model_id: str,
        layer_num: int,
        hook_type: str,
        vector: list[float],
        default_steer_strength: float | None = 10,
    ) -> NPVector:
        if hook_type not in VALID_HOOK_TYPES:
            raise ValueError(
                f"Invalid hook name: {hook_type}. Valid hook names are {VALID_HOOK_TYPES}."
            )
        payload = {
            "modelId": model_id,
            "layerNumber": layer_num,
            "hookType": hook_type,
            "vector": vector,
            "vectorDefaultSteerStrength": default_steer_strength,
            "vectorLabel": label,
        }
        response = self.send_request(
            method="POST",
            json=payload,
            uri="new",
        )

        return NPVector(
            model_id=response["vector"]["modelId"],
            source=response["vector"]["source"],
            index=response["vector"]["index"],
            label=response["vector"]["label"],
            hook_name=response["vector"]["hookName"],
            values=response["vector"]["values"],
            default_steer_strength=response["vector"]["defaultSteerStrength"],
            url=response["url"],
        )

    def delete(self, vector: NPVector) -> Response:
        return self._delete(vector.model_id, vector.source, vector.index)

    def _delete(self, modelId: str, source: str, index: str) -> Response:
        payload = {
            "modelId": modelId,
            "source": source,
            "index": index,
        }
        return self.send_request(
            method="POST",
            uri="delete",
            json=payload,
        )

    def get(self, model_id: str, source: str, index: str) -> NPVector:
        response = self.send_request(
            method="POST",
            uri="get",
            json={"modelId": model_id, "source": source, "index": index},
        )
        return NPVector(
            model_id=response["vector"]["modelId"],
            source=response["vector"]["layer"],
            index=response["vector"]["index"],
            label=response["vector"]["vectorLabel"],
            hook_name=response["vector"]["hookName"],
            values=response["vector"]["vector"],
            default_steer_strength=response["vector"]["vectorDefaultSteerStrength"],
        )

    def get_owned(self) -> list[NPVector]:
        response = self.send_request(
            method="POST",
            uri="list-owned",
        )
        return [
            NPVector(
                model_id=vector["modelId"],
                source=vector["layer"],
                index=vector["index"],
                label=vector["vectorLabel"],
                hook_name=vector["hookName"],
                values=vector["vector"],
                default_steer_strength=vector["vectorDefaultSteerStrength"],
            )
            for vector in response["vectors"]
        ]



import os
from typing import Optional

import requests
from dotenv import load_dotenv

load_dotenv()


class NPKeyMissingError(Exception):
    """No API key provided. Use context manager, global setting, parameter, or environment variable."""

    pass


class NPUnauthorizedError(Exception):
    """Unauthorized. Check your API key configuration."""

    pass


class NPRateLimitError(Exception):
    """Rate limit exceeded. Please try again later or email support@neuronpedia.org to raise your limit."""

    pass


class NPInvalidResponseError(Exception):
    """The API returned an invalid JSON response. Please contact support@neuronpedia.org."""

    pass


class NPRequest:
    USE_LOCALHOST = os.getenv("USE_LOCALHOST", "false").lower() == "true"
    BASE_URL = (
        "https://neuronpedia.org/api"
        if not USE_LOCALHOST
        else "http://localhost:3000/api"
    )

    def __init__(
        self,
        endpoint: str,
        api_key: Optional[str] = None,
    ):
        if not self.BASE_URL:
            raise ValueError("BASE_URL must be set in a subclass or globally.")
        self.endpoint = endpoint
        self.api_key = self._get_api_key(api_key)

    @staticmethod
    def _get_api_key(api_key: Optional[str] = None):
        """Retrieve and validate the NEURONPEDIA_API_KEY from parameter, context, or environment variable."""
        # 1. Use explicitly passed parameter if provided
        if api_key:
            return api_key

        # 2. Check context variable
        try:
            from neuronpedia import get_api_key

            context_api_key = get_api_key()
            if context_api_key:
                return context_api_key
        except ImportError:
            # Handle case where neuronpedia module isn't fully loaded yet
            pass

        # 3. Fall back to environment variable
        env_api_key = os.getenv("NEURONPEDIA_API_KEY")
        if not env_api_key:
            raise NPKeyMissingError(
                "No API key provided. Use one of these methods:\n"
                "1. Context manager: with neuronpedia.api_key('key'): ...\n"
                "2. Global setting: neuronpedia.set_api_key('key')\n"
                "3. Parameter: FeatureRequest(api_key='key')\n"
                "4. Environment: set NEURONPEDIA_API_KEY variable"
            )
        return env_api_key

    def get_url(
        self,
    ):
        """Construct the full URL for the request."""
        return f"{self.BASE_URL}/{self.endpoint}"

    def send_request(
        self,
        uri: str = "",
        method: str = "POST",
        **kwargs,
    ):
        """Send an HTTP request."""
        headers = kwargs.get(
            "headers",
            {},
        )
        headers["X-Api-Key"] = self.api_key
        headers["Accept-Encoding"] = "gzip"
        url = f"{self.BASE_URL}/{self.endpoint}/{uri}"
        print(f"Sending {method} request to {url}")
        # print(f"Body: {json.dumps(kwargs.get('json', {}), indent=2)}")

        response = requests.request(
            method,
            url,
            headers=headers,
            **kwargs,
        )

        if response.status_code == 401:
            raise NPUnauthorizedError(
                "Unauthorized. Check your API key using one of these methods:\n"
                "1. Context: with neuronpedia.api_key('key'): ...\n"
                "2. Global: neuronpedia.set_api_key('key')\n"
                "3. Parameter: Request(api_key='key')\n"
                "4. Environment: NEURONPEDIA_API_KEY variable"
            )
        elif response.status_code == 404:
            raise requests.exceptions.HTTPError(
                f"Resource not found: {response.status_code}"
            )
        elif response.status_code == 429:
            raise NPRateLimitError(
                "You've exceeded the API rate limit. Try later or email support@neuronpedia.org to raise your limit."
            )
        elif 400 <= response.status_code < 500:
            raise requests.exceptions.HTTPError(
                f"Request failed with status {response.status_code}: {response.json()}"
            )
        elif 500 <= response.status_code < 600:
            raise requests.exceptions.HTTPError(
                f"Server error occurred: {response.status_code}. Try again later or contact support@neuronpedia.org."
            )
        response.raise_for_status()  # Raise an exception for HTTP errors
        try:
            response_json = response.json()
            print("Got a successful response.")
            # print(json.dumps(response_json, indent=2))
            return response_json
        except requests.exceptions.JSONDecodeError:
            print("Error: Status code:", response.status_code)
            print(f"Error: Response text: {response.text[:1024]}")
            raise NPInvalidResponseError(
                "The API returned an invalid JSON response. Please contact support@neuronpedia.org."
            )



from dataclasses import dataclass
from typing import List

from neuronpedia.np_activation import Activation
from neuronpedia.np_explanation import Explanation


@dataclass
class Logit:
    token: str
    value: float


@dataclass
class Feature:
    modelId: str
    source: str
    index: int
    density: float | None = None
    top_logits: List[Logit] | None = None
    bottom_logits: List[Logit] | None = None
    explanations: List[Explanation] | None = None
    activations: List[Activation] | None = None

    @classmethod
    def get(cls, model_id: str, source: str, index: int) -> "Feature":
        from neuronpedia.requests.feature_request import FeatureRequest

        request = FeatureRequest()
        return request.get(model_id, source, index)

    def open_in_browser(self):
        import webbrowser

        webbrowser.open(f"https://neuronpedia.org/{self.modelId}/{self.source}/{self.index}")



import json
from typing import Optional

from neuronpedia.np_sae_feature import SAEFeature
from neuronpedia.requests.base_request import NPRequest


class SAEFeatureRequest(NPRequest):
    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        super().__init__("feature", api_key=api_key)

    def get(self, model_id: str, source: str, index: str) -> SAEFeature:
        result = self.send_request(method="GET", uri=f"{model_id}/{source}/{index}")

        return SAEFeature(
            modelId=result["modelId"],
            source=result["layer"],
            index=result["index"],
            jsonData=json.dumps(result),
        )



from typing import Optional

from neuronpedia.np_model import Model
from neuronpedia.requests.base_request import NPRequest


class ModelRequest(NPRequest):
    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        super().__init__("model", api_key=api_key)

    def new(
        self,
        id: str,
        layers: int,
        display_name: str | None = None,
        url: str | None = None,
    ) -> Model:
        """
        Create a new model on Neuronpedia. It will be UNLISTED, meaning that anyone who knows its ID will be
        able to view it, but it won't be listed publicly in the home page and dropdowns.
        Args:
            id: Unique identifier for the model (eg "gemma-2-2b")
            layers: Number of layers in the model (eg 26)
            display_name: Human-readable name for the model (optional, eg "Gemma 2 2B")
            url: URL associated with the model or its creator (optional, eg "https://deepmind.google/, or
                Huggingface URL")

        Returns:
            Model: Response containing the created model information
        """
        payload = {
            "id": id,
            "layers": layers,
            "displayName": display_name,
            "url": url,
        }
        response = self.send_request(
            method="POST",
            json=payload,
            uri="new",
        )
        return Model(
            id=response["id"],
            layers=response["layers"],
            display_name=response["displayName"],
            url=response["website"],
        )
